@@ -4,11 +4,11 @@ use strict;
 use Net::SSLeay;
 use POE qw (Filter::HTTPD Filter::Stackable Wheel::ReadWrite);
 use Scalar::Util qw(blessed);
-use Carp qw(carp);
+use Carp qw(carp confess);
 use POE;
 
 use vars qw($VERSION @ISA);
-$VERSION = '0.28';
+$VERSION = '0.30';
 sub DOSENDBACK () { 1 }
 
 our $globalinfos;
@@ -217,10 +217,15 @@ sub new {
    $self->{debug} = $params->{debug} || 0;
    $self->{cacrl} = $params->{cacrl} || undef;
    $self->{client} = $params->{client} || 0;
+   $self->{errorhandler} = $params->{errorhandler};
    $self->{params} = $params;
 
    $self->{context} = Net::SSLeay::CTX_new();
 
+   if ((!$self->{client}) && (!$params->{"nohonor"})) { # Beim Apache: SSLHonorCipherOrder
+      Net::SSLeay::CTX_set_options($self->{context}, 0x00400000); # SSL_OP_CIPHER_SERVER_PREFERENCE
+   }
+
    Net::SSLeay::CTX_use_RSAPrivateKey_file($self->{context}, $params->{key}, &Net::SSLeay::FILETYPE_PEM);
    Net::SSLeay::CTX_use_certificate_file($self->{context}, $params->{crt}, &Net::SSLeay::FILETYPE_PEM);
    if ($params->{chain}) {
@@ -243,14 +248,25 @@ sub new {
    $self->{ssl} = Net::SSLeay::new($self->{context});
    Net::SSLeay::set_bio($self->{ssl}, $self->{rbio}, $self->{wbio});
 
-   if ($params->{dhcert}) {
-      my $dhbio = Net::SSLeay::BIO_new_file($params->{dhcert}, "r");
-      die "Cannot open dhcert file!" unless $dhbio;
+   if ($params->{dhcert} ||
+       $params->{dhcertmem}) {
+      my $dhbio = undef;
+      if ($params->{dhcertmem}) {
+         $dhbio = Net::SSLeay::BIO_new(Net::SSLeay::BIO_s_mem());
+         my $sent = Net::SSLeay::BIO_write($dhbio, $params->{dhcertmem});
+         die "Cannot write to dhcert bio!"
+            if (($sent) != length($params->{dhcertmem}));
+      } else {
+         die "Cannot open dhcert file!"
+            unless (-f $params->{dhcert} && ($dhbio = Net::SSLeay::BIO_new_file($params->{dhcert}, "r")));
+      }
       my $dhret = Net::SSLeay::PEM_read_bio_DHparams($dhbio);
       Net::SSLeay::BIO_free($dhbio);
-      die "Couldn't set DH parameters!" if (Net::SSLeay::set_tmp_dh($self->{ssl}, $dhret) < 0);
+      die "Couldn't set DH parameters!"
+         if (Net::SSLeay::set_tmp_dh($self->{ssl}, $dhret) < 0);
       my $rsa = Net::SSLeay::RSA_generate_key(2048, 73);
-      die "Couldn't set RSA key!" if (!Net::SSLeay::CTX_set_tmp_rsa($self->{context}, $rsa));
+      die "Couldn't set RSA key!"
+         if (!Net::SSLeay::CTX_set_tmp_rsa($self->{context}, $rsa));
       Net::SSLeay::RSA_free($rsa);
    }
 
@@ -423,13 +439,39 @@ sub doSSL {
          $self->{accepted}++;
          $ret++;
       } else {
+         my $errtext = $!;
          my $err2 = Net::SSLeay::get_error($self->{ssl}, $err);
          unless ($err2 == Net::SSLeay::ERROR_WANT_READ()) {
-            carp("POE::Filter::SSL: UNEXPECTED ERROR: ERR1:".$err." ERR2:".$err2.($self->{client} ? '' : " HINT: ".
-                "Check if you have configured a CRT and KEY file, and that ".
-                "both are readable")); # unless ($err2 == 5); # SSL_ERROR_SYSCALL
-            $ret++ unless $self->{accepted}++;
-            return $ret;
+            my $tmp = "POE::Filter::SSL: ".($self->{client} ? "connect" : "accept").": ";
+            my $err3 = undef;
+            if ($err3 = Net::SSLeay::ERR_get_error()) {
+               $tmp .= Net::SSLeay::ERR_error_string($err3)."(".$err3.", ".$err2.")";
+            } else {
+               $tmp .= "No error (return=".$err2.")";
+            }
+            if (defined($self->{errorhandler})) {
+               if (ref($self->{errorhandler}) eq "CODE") {
+                  $self->{errorhandler}($self, {
+                     ssl       => $self->{ssl},
+                     msg       => $tmp,
+                     ret       => $err,
+                     get_error => $err2,
+                     error     => $err3,
+                  });
+               } elsif(lc($self->{errorhandler}) eq "ignore") {
+               } elsif(lc($self->{errorhandler}) eq "carp") {
+                  carp($tmp);
+               } elsif(lc($self->{errorhandler}) eq "confess") {
+                  confess($tmp);
+               } elsif(lc($self->{errorhandler}) eq "carponetime") {
+                  carp($tmp)
+                     unless $self->{errorstat}->{$err||"-"}->{$err2||"-"}->{$err3||"-"}++;
+               }
+            } else {
+               carp($tmp);
+            }
+            $ret++
+               unless $self->{accepted}++;
          }
       }
    }
@@ -525,11 +567,11 @@ POE::Filter::SSL - The easiest and flexiblest way to SSL in POE!
 
 =head1 VERSION
 
-Version 0.28
+Version 0.30
 
 =head1 DESCRIPTION
 
-This module allows to secure connections of I<POE::Wheel::ReadWrite> with OpenSSL by a
+This module allows one to secure connections of I<POE::Wheel::ReadWrite> with OpenSSL by a
 I<POE::Filter> object, and behaves (beside of SSLing) as I<POE::Filter::Stream>.
 
 I<POE::Filter::SSL> can be added, switched and removed during runtime, for example if you
@@ -553,7 +595,7 @@ Server and client mode
 
 Optional client certificate verification
 
-Allows to accept connections with invalid or missing client certificate and return custom error data
+Allows one to accept connections with invalid or missing client certificate and return custom error data
 
 CRL check of client certificates
 
@@ -673,8 +715,9 @@ By default I<POE::Filter::SSL> acts as a SSL server. To use it in client mode yo
 =item Advanced Example
 
 This example is an IMAP-Relay which forwards the connections to a IMAP server
-by username. It allows the uncrypted transfer on port 143, with the option
-of SSL on the established connection (STARTTLS). On port 993 it allows to do direct SSL.
+by username. It allows one the uncrypted transfer on port 143, with the option
+of SSL on the established connection (STARTTLS). On port 993 it allows one to do
+direct SSL.
 
 Tested with Thunderbird version 3.0.5.
 
@@ -955,18 +998,26 @@ In this case, you normaly have no I<key> and I<crt> option.
 
 Configures a CRL (ca.crl) against the client certificate is verified by I<clientCertValid()>.
 
-=item dhcert
+=item dhcert{mem}
 
 If you want to enable perfect forward secrecy, here you can enable Diffie-Hellman. You just have to create a dhparam file and there here the path to the path/to/FILENAME.pem where your Diffie-Hellman (pem format) stays.
 
 openssl dhparam -check -text -5 2048 -out path/to/FILENAME.pem
 
+You are able to pass the already inmemory dhparam file as scalar(string) via I<dhcertmem>.
+
 =item clientcert
 
 Only in server mode: Request during ssl handshake from the client a client certificat.
 
 B<WARNING:> If the client provides an untrusted or no client certficate, the connection is B<not> failing. You have to ask I<clientCertValid()> if the certicate is valid!
 
+=item nohonor
+
+By default, as server, I<POE::Filter:SSL> sets the option I<SSL_OP_CIPHER_SERVER_PREFERENCE>. For more information you may google the pendant of apache I<SSLHonorCipherOrder>.
+
+To flip back to the old behaviour, not setting this option, you can set nohonor.
+
 =item cipher
 
 Specify which ciphers are allowed for the synchronous encrypted transfer of the data over the ssl connection.
@@ -1031,7 +1082,7 @@ also tests against the CRL, if you have the I<cacrl> option set on I<new()>.
 
 B<WARNING:> POE::Filter:SSL now is able to do the ssh handshake now without any helpers. Because of this, this function has been removed!
 
-Allows to add filters after the ssl handshake. It has to be called in the input handler, and needs the passing of the I<POE::Wheel::ReadWhile> object. If it returns false, you have to return from the input handler.
+Allows one to add filters after the ssl handshake. It has to be called in the input handler, and needs the passing of the I<POE::Wheel::ReadWhile> object. If it returns false, you have to return from the input handler.
 
 See the I<HTTPS-Server>, I<SSL on an established connection> and I<Client certificate verification> examples in I<SYNOPSIS>
 
@@ -1041,6 +1092,44 @@ Returns I<true> if there is a client certificate, that might be untrusted.
 
 B<WARNING:> If the client provides an untrusted client certficate a client certicate that is listed in CRL, this function returns I<true>. You have to ask I<clientCertValid()> if the certicate is valid!
 
+=item errorhandler
+
+By default, every ssl error is escalated via carp. You may change this behaviour via this option to:
+
+=over 2
+
+=item "ignore"
+
+Do not report any error.
+
+=item I<CODE>
+
+Setting errorhandler to a reference of a function allows to be called it callback function with the following options:
+
+ARG1: POE:SSL::Filter instance
+
+ARG2: Ref on a Hash with the following keys:
+
+  ret        The return code of Net::SSLeay::connect (client) or Net::SSLeay::accept (server)
+  ssl        The SSL context (SSL_CTX)
+  msg        The error message as text, as normaly reported via carp
+  get_error  The error code of get_error the ssl context
+  error      The error code of get_error without context
+
+=item "carp" (or undef)
+
+Do Carp/carp on error.
+
+=item "confess"
+
+Do Carp/confess (stacktrace) on error.
+
+=item "carponetime"
+
+Report carp for one occurence only one time - over all!
+
+=back
+
 =item debug
 
 Shows debug messages of I<clientCertNotOnCRL()>.
@@ -1130,11 +1219,13 @@ L<http://search.cpan.org/dist/POE-Filter-SSL>
 
 =head1 Commercial support
 
-Commercial support can be gained at <sslsupport at priv.de>
+Commercial support can be gained at <sslsupport at priv.de>.
+
+Used in our products, you can find on L<https://www.cryptomagic.eu/>
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2010-2014 Markus Schraeder, all rights reserved.
+Copyright 2010-2016 Markus Schraeder, CryptoMagic GmbH, all rights reserved.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
